#include <cuda.h>
#include <dlfcn.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#define LIBCUDA "libcuda.so.1"

#define TIMEOUT_DEFAULT 30
#define TIMEOUT_MAX     300

#ifdef DEBUG
#define CHK(t, err, err_rv) do {             \
      fprintf(stderr, "[" #t "] running\n"); \
      bool rv = (t); \
      fprintf(stderr, "[" #t "] returns %s\n", rv?"true":"false"); \
      if (!rv) {                            \
        fprintf(stderr, "ERROR: " err "\n"); \
        exit(err_rv);                        \
      }                                      \
    } while(0)
#else
#define CHK(t, err, err_rv) do {             \
      if (!(t)) {                            \
        fprintf(stderr, "ERROR: " err "\n"); \
        exit(err_rv);                        \
      }                                      \
    } while(0)
#endif

#define CUDAERR(func, err) do {              \
      const char *name, *str;                \
      if ((*p_cuGetErrorName)((err), &name) != CUDA_SUCCESS) {  \
        name = "";                           \
      }                                      \
      if ((*p_cuGetErrorString)((err), &str) != CUDA_SUCCESS) { \
        str = "Unknown error!";              \
      }                                      \
      /* Print error str w/ upper-case first letter,
         allowing for str's with zero or one chars */ \
      printf("%c%s (%s) %s=%d\n", toupper(*str), (*str)?(str+1):"", name, (func), (err)); \
    } while(0)

void handle_timeout(int signum)
{
    printf("CUDA Driver initialization timed out!\n");
    exit(1);
}

int main(int argc, const char **argv)
{
    typedef CUresult (*cuInit_t)(unsigned int Flags);
    typedef CUresult (*cuDevicePrimaryCtxRetain_t)(CUcontext* pctx, CUdevice dev);
    typedef CUresult (*cuGetErrorName_t)(CUresult error, const char** pStr);
    typedef CUresult (*cuGetErrorString_t)(CUresult error, const char** pStr);

    cuInit_t p_cuInit = NULL;
    cuDevicePrimaryCtxRetain_t p_cuDevicePrimaryCtxRetain = NULL;
    cuGetErrorName_t p_cuGetErrorName = NULL;
    cuGetErrorString_t p_cuGetErrorString = NULL;

    void *libcuda = NULL;

    // set timeout
    signal(SIGALRM, handle_timeout);
    int timeout = TIMEOUT_DEFAULT;
    const char *timeout_env = NULL;
    if ((timeout_env = getenv("_CUDA_COMPAT_TIMEOUT")) != NULL) {
        int timeout_new = atoi(timeout_env);
        if ((timeout_new >= 0) && (timeout_new <= TIMEOUT_MAX)) {
            timeout = timeout_new;
#ifdef DEBUG
            fprintf(stderr, "Setting %d second timeout (env setting was '%s')\n", timeout, timeout_env);
#endif
        }
#ifdef DEBUG
        else {
            fprintf(stderr, "Ignoring invalid timeout setting '%s'\n", timeout_env);
        }
#endif
    }
#ifdef DEBUG
    else {
        fprintf(stderr, "Setting default %d second timeout\n", timeout);
    }
#endif
    if (timeout) {
        alarm(timeout);
    }

    // open libcuda and fetch function pointers
    CHK(((libcuda = dlopen(LIBCUDA, RTLD_LAZY)) != NULL), "Couldn't dlopen() CUDA Driver", 2);
    CHK(((p_cuInit = (cuInit_t)dlsym(libcuda, "cuInit")) != NULL), "Couldn't find cuInit() in libcuda", 3);
    CHK(((p_cuDevicePrimaryCtxRetain = (cuDevicePrimaryCtxRetain_t)dlsym(libcuda, "cuDevicePrimaryCtxRetain")) != NULL), "Couldn't find cuDevicePrimaryCtxRetain() in libcuda", 3);
    CHK(((p_cuGetErrorName = (cuGetErrorName_t)dlsym(libcuda, "cuGetErrorName")) != NULL), "Couldn't find cuGetErrorName() in libcuda", 3);
    CHK(((p_cuGetErrorString = (cuGetErrorString_t)dlsym(libcuda, "cuGetErrorString")) != NULL), "Couldn't find cuGetErrorString() in libcuda", 3);

    // try initializing CUDA
    CUresult res;
#ifdef DEBUG
    CHK(((res = (*p_cuInit)(0)) == CUDA_SUCCESS), "cuInit failed", 5);
    if (res != CUDA_SUCCESS)
#else
    if ((res = (*p_cuInit)(0)) != CUDA_SUCCESS)
#endif
    {
        CUDAERR("cuInit()", res);
        exit(1);
    }
    CUcontext ctx;
#ifdef DEBUG
    CHK(((res = (*p_cuDevicePrimaryCtxRetain)(&ctx, 0)) == CUDA_SUCCESS), "context create failed", 6);
    if (res != CUDA_SUCCESS)
#else
    if ((res = (*p_cuDevicePrimaryCtxRetain)(&ctx, 0)) != CUDA_SUCCESS)
#endif
    {
        CUDAERR("cuDevicePrimaryCtxRetain()", res);
        exit(1);
    }

    CHK((dlclose(libcuda) == 0), "dlclose() failed", 4);

    // NOTE: Do not change the text of this message; entrypoint
    // scripts look for the exact string "CUDA Driver OK"
    printf("CUDA Driver OK\n");
    return 0;
}
